require 'common.hsv'
require 'common.util'
require 'common.glutil'
require 'common.fbo'
require 'common.shader'

glush.gui.checkbox('useShader',true)
glush.gui.slider('maxiter', 0, 40, 100)
local hsvTexDetail = 1024
local volDim = 64

local hsvTex = HsvTex.new(hsvTexDetail)

local volShader = ShaderProgram.new{
	vertexCode=[[
varying vec3 v;
varying vec3 eye;
void main() {
	//this only needs to be done once per render
	//i.e. uniform?
	eye = (gl_ModelViewMatrixInverse * vec4(0., 0., 0., 1.)).xyz;
	v = gl_Vertex.xyz;	//[-1,1]^3
	gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;
}
]];
	fragmentCode=[[
varying vec3 v;
varying vec3 eye;
uniform sampler3D volTex;
uniform float maxiter;
void main() {
	vec3 p = v;
	vec4 result = vec4(0., 0., 0., 1.);
	
	vec4 texel = texture3D(volTex, p*.5+.5);
	result.rgb += result.a * texel.a * texel.rgb;
	result.a *= 1. - texel.a;
	
	vec3 step = v - eye;
	step = normalize(step) * 2. / maxiter;
	for (float i = 2.; i <= maxiter; i++) {
		p += step;
		if (p.x < -1. || p.y < -1. || p.z < -1. ||
			p.x > 1. || p.y > 1. || p.z > 1.) break;
		
		//notice if you store 1-alpha in the color you trace through the volume
		//then you can forward-trace your blending
		//instead of having to backward-trace
		//(as you would when rendering transparent stuff on top of each other)
		//this will allow you to bailout early if your transparency ever hits fully opaque
		texel = texture3D(volTex, p*.5+.5);
		result.rgb += result.a * texel.a * texel.rgb;
		result.a *= 1. - texel.a;
		
		if (result.a == 0.) break;
	}
	gl_FragColor = vec4(result.rgb, 1. - result.a);
}
]];
	uniforms={volTex=0};
}

-- volume creation

local volGenShader = ShaderProgram.new{
	vertexCode=[[
varying vec2 tc;
void main() {
	tc = gl_Vertex.xy;
	gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;
}
]];
	fragmentCode=[[
varying vec2 tc;
uniform float slice;
uniform sampler1D hsvTex;
uniform vec4 c0;
uniform vec4 zmin, zmax;
#define MAXITER 10.
void main() {
	vec4 u = vec4(.5, tc.x, tc.y, slice);		//[0,1]^4
	vec4 c = c0;
	vec4 z = u * (zmax - zmin) + zmin;
	
	gl_FragColor = vec4(0., 0., 1., 1.);
	
	float i = 0.;
	for (; i < MAXITER; i++) {
		float dt = dot(z.xyz, z.xyz);
		z = vec4(2. * z.w * z.xyz, z.w * z.w - dt) + c;
		if (dot(z,z) > 4.) break;
	}
	if (i < MAXITER) {
		//TODO - log adjust color to get fractional iteration
		//TODO - store iteration in the volume data
		//then do coloring (and lighting) in the render (raytrace) shader
		gl_FragColor = texture1D(hsvTex, (i + .5) / 50.);
		gl_FragColor.a = .02;
	}
}
]];
	uniforms={
		hsvTex = 0;
		zmin = {-2,-2,-2,-2};
		zmax = { 2, 2, 2, 2};
		--c0 = {0, 0, 0, 0};
		c0 = {-0.213,-0.0410,-0.563,-0.560};
	}
}

-- volume texture
local volTex = Tex3D.new{
	wrap={s=gl.GL_CLAMP, t=gl.GL_CLAMP, r=gl.GL_CLAMP};
	minFilter=gl.GL_LINEAR;
	magFilter=gl.GL_LINEAR;
	internalFormat=gl.GL_RGBA;
	width=volDim;
	height=volDim;
	depth=volDim;
}
volTex:unbind()

-- fbo init
local genFbo = FBO.new()

function regenVolume()
	glreport('before regen volume')
	gl.glPushAttrib(gl.GL_ENABLE_BIT, gl.GL_VIEWPORT_BIT)
	gl.glDisable(gl.GL_DEPTH_TEST)
	-- set up viewport and matrix
	gl.glViewport(0, 0, volDim, volDim)
	drawOverlay(function()
		
		volGenShader:use()
		-- update zmin, zmax, c0
		hsvTex:bind()
		glreport('bind vol gen shader')

		for z=0,volDim-1 do
			-- frame callback
			genFbo:setColorAttachmentTex3D(0, volTex.id, z)
			genFbo:bind()
			assert(genFbo:check())
			glreport('framebuffer texture3d slice '..z)
			-- do the render here
			gl.glUniformf('slice',(z+.5)/volDim)	-- from 0 to 1
			gl.glBegin(gl.GL_QUADS)
			gl.glVertex(0,0)
			gl.glVertex(1,0)
			gl.glVertex(1,1)
			gl.glVertex(0,1)
			gl.glEnd()
			genFbo:unbind()
		end
		
		hsvTex:unbind()
		ShaderProgram.useNone()
	end)
	
	gl.glPopAttrib()
	glreport('after regen volume')
end

regenVolume()

function vtxForBits(i)
	return
		glush.bitand(i-1,1)*2-1,
		glush.bitand(i-1,2)-1,
		glush.bitand(i-1,4)*.5-1
end

local vtxs = {}
for dim=1,8 do
	table.insert(vtxs, {vtxForBits(dim)})
end

local faces = {
	{f={1,2,4,3}, n={0,0,-1}};
	{f={5,6,2,1}, n={0,-1,0}};
	{f={7,8,6,5}, n={0,0,1}};
	{f={3,4,8,7}, n={0,1,0}};
	{f={7,5,1,3}, n={-1,0,0}};
	{f={6,8,4,2}, n={1,0,0}};
}

gl.glEnable(gl.GL_DEPTH_TEST)
gl.glEnable(gl.GL_CULL_FACE)
function glush.display()

	if useShader then
		volShader:use()
		gl.glUniformf('maxiter', maxiter)
		gl.glFrontFace(gl.GL_CW)
	end

	gl.glEnable(gl.GL_TEXTURE_3D)
	volTex:bind()
	if cubeList then
		gl.glCallList(cubeList)
	else
		cubeList = gl.glGenList()
		gl.glNewList(cubeList, gl.GL_COMPILE_AND_EXECUTE)
		-- todo - display list
		gl.glBegin(gl.GL_QUADS)
		for _,f in ipairs(faces) do
			gl.glNormal(f.n)
			for _,i in ipairs(f.f) do
				local v = {vtxForBits(i)}
				if not useShader then
					-- todo - texgen?
					gl.glTexCoord(v[1]*.5+.5, v[2]*.5+.5, v[3]*.5+.5)
				end
				gl.glVertex(v)
			end
		end
		gl.glEnd()
		gl.glEndList()
	end
	volTex:unbind()
	gl.glDisable(gl.GL_TEXTURE_3D)
	if useShader then
		ShaderProgram.useNone()
		gl.glFrontFace(gl.GL_CCW)
	end
end

